

#include <types.h>
#include <setup.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <network.h>
#include <hardware.h>
#include <netdev.h>
#include <time.h>
#include <cs8900.h>

#define CS8900_BASE_ADDR		(0x10000300)

#define report(fmt, args...)	printf(fmt, ##args)
#define assert(fmt, args...)	printf(fmt, ##args)
#define failed(fmt, args...)	printf(fmt, ##args)

static bool cs8900_init(void);
static void cs8900_exit(void);
static int cs8900_recv_packet(void);
static bool cs8900_send_packet(void *packet, int len);
static void cs8900_enable(void);
static void cs8900_disable(void);

static void write_to_ppr(uint addr, uint16 value);
static uint16 read_from_ppr(uint addr);

static const unsigned long base = CS8900_BASE_ADDR;

extern bool eth_init(void){
	return cs8900_init();
}

extern void eth_exit(void){
	cs8900_exit();
	return;
}

extern bool eth_recv_poll(void){
	return cs8900_recv_packet();
}

extern bool eth_send(void *packet, int length){
	return cs8900_send_packet(packet, length);
}

extern void eth_enable(void){
	cs8900_enable();
	return;
}

extern void eth_disable(void){
	cs8900_disable();
	return;
}

extern void eth_set_haddr(const void *haddr){
	const uint16 *mac = haddr;
	write_to_ppr(PP_IA+0, *mac++);
	write_to_ppr(PP_IA+2, *mac++);
	write_to_ppr(PP_IA+4, *mac++);
	return;
}

static bool cs8900_init(void){
	int i;
	uint16 *mac;

	/* bugfix */
	pxa_gpio_mode(GPIO15_nCS_1_MD);
	memset((void *)0x10000000, 0x0, 2);
	memset((void *)0x10000001, 0x0, 2);
	memset((void *)0x10000000, 0x0, 2);
	mdelay(1);

	// check memory mapping (read IOBase Address).
	// check value of IO_PPPTR is 0x3000.
	// 1~3 bit(011b) is read only. so clear others.
	if ((inw(base + ADD_PORT) & 0x7000) != 0x3000){
		failed(" can't access to memory of cs8900a.\n");
		return false;
	}
	
	// check cs8900a chip id (read packet page 0x0000~0x0003).
	if (read_from_ppr(PP_ChipID) != CHIP_EISA_ID_SIG){
		failed(" ethernet chip is not cs8900a. are you use cs8900a.\n");
		return false;
	}

	report(" cs8900a chip is at 0x%08lx\n", base);

	write_to_ppr(PP_LineCTL, 0x00);							// rx, tx disable

	write_to_ppr(PP_CS8900_ISAINT, INTRQ_HIGH_IMPEDENCE);	// interrupt is not used
	write_to_ppr(PP_CS8900_ISADMA, DMRQ_HIGH_IMPEDENCE);	// dma is not used

	// set setup and control registers
	write_to_ppr(PP_RxCFG, RX_OK_ENBL | RX_CRC_ERROR_ENBL);
	write_to_ppr(PP_RxCTL, RX_OK_ACCEPT | RX_IA_ACCEPT);
	
	write_to_ppr(PP_TxCFG,	TX_LOST_CRS_ENBL | TX_SQE_ERROR_ENBL | TX_OK_ENBL |
			TX_LATE_COL_ENBL | TX_JBR_ENBL | TX_ANY_COL_ENBL);
	write_to_ppr(PP_TxCMD, TX_START_ALL_BYTES);
	write_to_ppr(PP_BufCFG, 0x0000);
	write_to_ppr(PP_BusCTL, ENABLE_IRQ | IO_CHANNEL_READY_ON);
	write_to_ppr(PP_TestCTL, 0x0000);

	write_to_ppr(PP_TxCommand, 0x00c0);						// start tx after the entire frame is in the CS8900A.

	for (i=0; i < 4; i++)									// set Logical Address Filter
		write_to_ppr(PP_LAF+i*2, 0xffff);

	mac = (uint16 *)setup->myhaddr;
	write_to_ppr(PP_IA+0, *mac++);
	write_to_ppr(PP_IA+2, *mac++);
	write_to_ppr(PP_IA+4, *mac++);

	write_to_ppr(PP_LineCTL, SERIAL_RX_ON | SERIAL_TX_ON);	// rx, tx enable
	//cs8900_disable();
	return true;
}

static void cs8900_exit(void){
	write_to_ppr(PP_LineCTL, 0x00);							// rx, tx disable
}

static bool cs8900_send_packet(void *packet, int len){
	if (!(read_from_ppr(PP_LineST) & LINK_OK)){ printf(" send failed : not linked.\n"); return false; }

	outw(TX_AFTER_ALL, base + TX_CMD_PORT);						// write command
	outw(len, base + TX_LEN_PORT);								// write length
	while ((read_from_ppr(PP_BusST) & READY_FOR_TX_NOW)==0);	// ready?

	if (len & 0x01) len++;										// align 16bit
	outsw(base + TX_FRAME_PORT, packet, (len >> 1));			// write packet data
	if (len & 0x01) outw(*((uint16 *)packet + len - 1), base + TX_FRAME_PORT);
	return true;
}

static int cs8900_recv_packet(void){
	uint isq, setup, len, cnt;
	uchar _rxbuff[1024], *rxbuff=_rxbuff+2;

	while ((isq = inw(base + ISQ_PORT))){						// loop while something exists in ISQ
		if (isq & ISQ_RECEIVER_EVENT){							// some packet is received.
			setup = inw(base + RX_FRAME_PORT);					// 1. get setup
			if ((setup & RX_OK) && RX_IA_ACCEPT){
				len = inw(base + RX_FRAME_PORT);				// 2. get length
				cnt = len / 2 + len % 2;						// 3. get packet
				insw(base + RX_FRAME_PORT, rxbuff, cnt);
				netif_rx(rxbuff, len);							// 4. call packet handler
			} else {											// ignore invalid packet
			}
		} else if (isq & ISQ_TRANSMITTER_EVENT){				// ignore other events
		} else if (isq & ISQ_BUFFER_EVENT){
		} else if (isq & ISQ_RX_MISS_EVENT){
		} else if (isq & ISQ_TX_COL_EVENT){
		}
	}
	return true;
}

static void cs8900_enable(void){
	uint16 linectl;
	linectl = read_from_ppr(PP_LineCTL);
	linectl &= ~(SERIAL_RX_ON | SERIAL_TX_ON);
	write_to_ppr(PP_LineCTL, linectl);
	return;
}

static void cs8900_disable(void){
	uint16 linectl;
	linectl = read_from_ppr(PP_LineCTL);
	linectl |= SERIAL_RX_ON | SERIAL_TX_ON;
	write_to_ppr(PP_LineCTL, linectl);
	return;
}

static void write_to_ppr(uint addr, uint16 value){
	outw(addr, base + ADD_PORT);
	outw(value, base + DATA_PORT);
	return;
}

static uint16 read_from_ppr(uint addr){
	outw(addr, base + ADD_PORT);
	return inw(base + DATA_PORT);
}
